
; BIOS - Basic I/O System.

		NOLIST
		INCLUDE	"eZ80F91.inc"
		LIST

		ORG	bios					; Origin of BIOS.

; BIOS jump table.

boot:		jp	b_boot					;  0 Initialize.
wboot:		jp	b_wboot					;  1 Warm boot.
const:		jp	b_const					;  2 Console status.
conin:		jp	b_conin					;  3 Console input.
conout:		jp	b_conout				;  4 Console output.
listd:		jp	b_list					;  5 List output.
punch:		jp	b_punch					;  6 Punch output.
reader:		jp	b_reader				;  7 Reader input.
home:		jp	b_home					;  8 Home disk.
seldsk:		jp	b_seldsk				;  9 Select disk.
settrk:		jp	b_settrk				; 10 Select track.
setsec:		jp	b_setsec				; 11 Select sector.
setdma:		jp	b_setdma				; 12 Set DMA address.
read:		jp	b_read					; 13 Read 128 bytes.
write:		jp	b_write					; 14 Write 128 bytes.
listst:		jp	b_listst				; 15 List status.
sectrn:		jp	b_sectrn				; 16 Sector translate.

; BIOS subroutines start here.

; BOOT Routine.

; The  BOOT  entry  point  gets  control  from  the cold  start  loader  and  is  responsible  for  basic system
; initialization, including sending a sign-on message, which can be omitted in the first version. If the  IOBYTE
; function is implemented, it must be set at this point. The various system parameters that are set by the WBOOT
; entry point must be  initialized, and control is  transferred to the CCP at 0x3400 + b for further processing.
; Note that register C must be set to zero to select drive A.

b_boot:		di
		ld	sp,b_biosstack				; Set default stack.

; Set the alternate function for PORT D to use UART0.

		ld	a,01h
		out0	(PD_ALT2),a
		dec	a					; Save 1 byte over LD A,0.
		out0	(PD_ALT1),a
		ld	a,0FFh
		out0	(PD_DDR),a

; Configure UART0 for 115200,8,1,n.

		ld	a,00h
		out0	(UART0_FCTL),a				; Disable UART0 FIFO.
		ld	a,80h
		out0	(UART0_LCTL),a				; Enable access to BRG.
		ld	a,1Bh
		out0	(UART0_BRG_L),a				; Load low byte of BRG.
		ld	a,00h
		out0	(UART0_BRG_H),a				; Load high byte of BRG.
		ld	a,03h
		out0	(UART0_LCTL),a				; Select 8 bits, no parity, 1 stop.

		ld	hl,b_bootmsg				; Address of sign on message text.
		call	b_TXstringHL				; Display CP/M sign on message.

; Initialise CP/M IObyte.

		ld	hl,0000h				; Clear I/O & drive bytes.
		ld	(iobyte),hl

		call	b_MMCInit				; Initialise MMC FLASH Memory Card.
		jr	nc,b_boot1
		ld	hl,b_InitFail				; Address of error message text.
		call	b_TXstringHL				; Display error message.
		jr	$					; Halt system.
b_boot1:	ld	hl,b_InitSucc				; Address of success message text.
		call	b_TXstringHL				; Display success message.

; WBOOT Routine.

; The WBOOT entry point gets control when a warm start occurs. A warm start is performed whenever a user program
; branches to  location 0x0000, or when the CPU is reset  from the front panel.  The CP/M system  must be loaded
; from the first two tracks of drive A  up to, but not including, the BIOS,  or CBIOS, if the user has completed
; the patch. System parameters must be initialized as follows:

; Location 0,1,2
;	Set to JP WBOOT for warm starts (0x0000: jp 0x4a03 + b).
; Location 3
;	Set initial value of IOBYTE, if implemented in the CBIOS.
; Location 4
;	High nibble = current user number, low nibble = current drive.
; Location 5,6,7
;	Set to JP BDOS, which is the primary entry point to CP/M for transient programs (0x0005: jp 0x3c06 + b).

; Refer to  Section 6.9  of the  CP/M Operating  System Manual  for complete  details of   page zero  use. Upon
; completion of  the initialization,   the WBOOT  program must  branch to  the CCP  at 0x3400 + b to restart the
; system. Upon entry  to the CCP,  register C is  set to the  drive to select  after  system initialization. The
; WBOOT routine  should read  location 4 in  memory,  verify that is a  legal drive, and pass  it to the CCP  in
; register C.

b_wboot:	di						; Disable interrupts.
		ld	sp,b_biosstack				; Set default stack.
		ld	hl,tpabuf				; Address of BIOS DMA buffer.
		ld	(b_dmaadd),hl
		ld	a,0C3h					; Opcode for 'jp'.
		ld	(ramstart),a				; Load at start of RAM.
		ld	hl,wboot				; Address of jump for a warm boot.
		ld	(ramstart + 01h),hl
		ld	(bdosen),a				; Opcode for 'jp'.
		ld	hl,bdos					; Address of jump for the BDOS.
		ld	(bdosen + 01h),hl
		ld	a,(usrdrv)				; Get current logged drive.
		cp	b_maxdrvs				; Must be between 0 and b_maxdrvs.
		jr	c,b_wboot1				; Carry if <= b_maxdrvs, so it's legal.
		xor	a					; Illegal drive so reset drive to 0.
		ld	(usrdrv),a				; Save new drive number (0).
b_wboot1:	ld	c,a					; Pass drive number in C.
		jp	ccp					; And start CP/M by jumping to the CCP.

; CONST Routine.

; You should sample  the status of  the currently assigned  console device and  return 0xff in  register A if  a
; character is ready to read and 0x00 in register A if no console characters are ready.

b_const:	in0	a,(UART0_LSR)				; Read UART status register of COM Port 1.
		bit	0,a					; Test character ready bit.
		jr	z,b_cs1					; If Z=1 then character not available.
		ld	a,0FFh					; Set character is available flag.
		ret						; Done.
b_cs1:		xor	a					; Set character not available flag.
		ret						; Done.

; CONIN Routine.

; The next console character is read into register A, and the parity bit (bit 7) is set to zero.  If no  console
; character is ready, wait until a character is typed before returning.

b_conin:	in0	a,(UART0_LSR)				; Read UART status register of COM Port 1.
		bit	0,a					; Test character ready bit and loop
		jr	z,b_conin				; until a character is available.
		in0	a,(UART0_RBR)				; Get character from receive buffer
		and	7Fh					; Strip the high bit (CP/M rules).
		ret						; Done.

; CONOUT Routine.

; The character is sent from register C to the console output device. The character is in ASCII, with high-order
; parity bit set to zero. You might want to include a time-out on a line-feed or carriage return, if the console
; device requires some time interval at the end of the  line (such as a TI Silent 700 terminal). You can  filter
; out control  characters that  cause the  console device  to react  in a  strange way  (CTRL-Z causes  the Lear
; - Siegler terminal to clear the screen, for example).

b_conout:	push	af
b_co1:		in0	a,(UART0_LSR)				; Read UART status register of COM port 1.
		bit	5,a					; Test TX Data Register Empty ready bit and
		jr	z,b_co1					; loop until TX buffer is empty.
		pop	af
		out0	(UART0_THR),c				; TX the character.
		ret						; Done.

; LIST Routine.

; The character is sent from register C to the currently assigned listing device. The character is in ASCII with
; zero parity bit.

b_list:		ret					; Done.

; PUNCH Routine.

; The character is sent from register C to the  currently assigned punch device. The character is in ASCII  with
; zero parity.

b_punch:	ret					; Done.

; READER Routine.

; The next character is read  from the currently assigned reader  device into register A with  zero parity (high
; -order bit must be zero); an End-Of-File condition is reported by returning an ASCII CTRL-Z (0x1a).

b_reader:	ld	a,1Ah					; No reader so just fake an EOF and return.
		ret						; Done.

; SELDSK Routine.

; The disk drive given by register C is selected  for further operations, where register C contains 0 for  drive
; A, 1  for drive  B, and  so on  up to  15 for  drive P  (the standard  CP/M distribution version supports four
; drives). On each disk  select, SELDSK must return  in HL the base  address of a 16 byte  area, called the Disk
; Parameter Header, described  in Section 6.10  of the CP/M  Operating System Manual.  For standard floppy disk
; drives, the contents of the header and associated tables do not change; thus, the program segment included  in
; the sample CBIOS performs this operation automatically.

; If there  is an  attempt to  select a  nonexistent drive,  SELDSK  returns HL = 0x0000 as  an error indicator.
; Although SELDSK must return  the header address on  each call, it is  advisable to postpone the  physical disk
; select operation until  an I/O function  (seek, read, or  write) is actually  performed, because disk  selects
; often occur without ultimately performing any disk I/O,  and many controllers unload  the head of the  current
; disk before  selecting the  new drive.  This causes  an excessive  amount of  noise and  disk wear.  The least
; significant bit of register E is zero if this is the first occurrence of the drive select since the last  cold
; or warm start.

b_seldsk:	ld	hl,0000h				; Load HL with a return error code.
		ld	a,c					; Load A with disk to select.
		cp	b_maxdrvs				; Check if valid drive number. Must be between 0 and b_maxdrvs - 1.
		ret	nc					; Return with error if no carry. Drive is > b_maxdrvs.
		ld	(b_diskno),a				; Save it for later BIOS use.
		ld	l,a					; Disk number in range so compute DPH address. L = disk number 0, 1, 2, .. 15.
		add	hl,hl					; *  2.
		add	hl,hl					; *  4.
		add	hl,hl					; *  8.
		add	hl,hl					; * 16.
		ld	de,b_dpbase				; Base address of disk parameter block.
		add	hl,de					; HL = (diskno * 16) + dpbase.
		ret						; Done.

; HOME Routine.

; The disk head  of the currently  selected disk (initially  disk A) is  moved to the  track 0 position. If  the
; controller allows access to the  track 0 flag from the  drive, the head is stepped  until the track 0 flag  is
; detected. If the controller does not support this feature,  the HOME call is translated into a call to  SETTRK
; with a parameter of 0.

b_home:		ld	bc,0000h				; Select track 0. Fall through to b_settrk.

; SETTRK Routine.

; Register BC contains the track number for subsequent disk accesses on the currently selected drive. The sector
; number in BC  is the same  as the number  returned from the  SECTRAN entry point.  You can choose  to seek the
; selected track at this time or  delay the seek until the next  read or write actually occurs. Register  BC can
; take on values in the range 0 - 76 corresponding to valid track numbers  for standard floppy disk drives and 0
; - 65535 for nonstandard disk subsystems.

b_settrk:	ld	(b_track),bc				; Set track passed from BDOS in register BC.
		ret						; Done.

; SECSEC Routine.

; Register BC contains the sector number, 1 through  26, for subsequent disk accesses on the currently  selected
; drive. The sector number in BC is the same as the number returned from the SECTRAN entry point. You can choose
; to send this  information to the  controller at this  point or delay  sector selection until  a read or  write
; operation occurs.

b_setsec:	ld	(b_sector),bc				; Set sector passed from BDOS in register BC.
		ret						; Done.

; SETDMA Routine.

; Register  BC contains  the DMA  (Disk Memory  Access) address  for subsequent  read or  write operations.  For
; example, if B = 0x00 and C = 0x80 when  SETDMA is called, all subsequent read operations read their data  into
; 0x80 through 0xff and all  subsequent write operations get their  data from 0x80 through 0xff,  until the next
; call  to SETDMA occurs.  The initial DMA address  is assumed  to be  0x0080. The controller need  not actually
; support Direct Memory Access. If,  for example, all data transfers   are through I/O ports, the  CBIOS that is
; constructed uses the 128 byte  area  starting at the selected   DMA address for the  memory  buffer during the
; subsequent read or
; write operations.

b_setdma:	ld	(b_dmaadd),bc				; Set DMA address given by registers BC.
		ret						; Done.

; Read/Write setup Routine.

; Creates an MMC compatible sector buffer address from the track and sector number passed from CP/M.

b_rwsetup:	ld	de,(b_sector)
		ld	hl,(b_track)				; MMCSector = (b_track * 32) + b_sector.
		add	hl,hl					; * 2
		add	hl,hl					; * 4
		add	hl,hl					; * 8
		add	hl,hl					; * 16
		add	hl,hl					; * 32
		add	hl,de					; Add in sector.
		ex	de,hl					; Put sector number in DE.
		ld	hl,0000h				; Zero HL for now.
		ret

; READ Routine.

; Assuming the drive has been selected, the track has been set, and the DMA address has been specified, the READ
; subroutine attempts to read one sector based upon  these parameters and returns the following error codes   in
; register A:

; 0 - no errors occurred.
; 1 - nonrecoverable error condition occurred.

; Currently, CP/M responds only to a zero or nonzero value as the return code. That is, if the value in register
; A is 0x00, CP/M assumes that the disk operation was completed properly.  If an error  occurs the CBIOS  should
; attempt at least 10 retries to see if the error is recoverable. When an error is reported the BDOS prints  the
; message BDOS ERR ON x: BAD SECTOR.  The operator then has the  option of pressing a carriage return  to ignore
; the error, or CTRL-C to abort.

b_read:		call	b_rwsetup
		call	b_MMCReadSec
		ld	hl,b_dbuff				; Get MMC sector buffer.
		ld	de,(b_dmaadd)				; BIOS sector buffer.
		ld	bc,0080h				; Move a total of 128 bytes.
		ldir						; Move data.
		jr	nc,b_w2
		jr	b_w1					; Return error condition.

; WRITE Routine.

; Data is written from the  currently selected DMA address to  the currently selected drive, track,  and sector.
; For floppy  disks, the  data should  be marked  as non deleted data to  maintain compatibility with other CP/M
; systems. The error codes given in the READ command are returned in register A, with error recovery attempts as
; described above.

; On each call to WRITE, the BDOS provides the following information in register C:

; 0x00 (normal sector write).
; 0x01 (write to directory sector).
; 0x02 (write to the first sector of a new data block).

; Condition 0 occurs whenever the next write operation is into a previously written area, such as a random  mode
; record update; when the write is to other than the first sector of an unallocated block; or when the write  is
; not into the directory area. Condition 1 occurs when a write into the directory area is performed. Condition 2
; occurs when the first  record (only) of a  newly allocated data block  is written. In most  cases, application
; programs read  or write  multiple 128 byte  sectors in  sequence; thus,  there is  little overhead involved in
; either operation when blocking and deblocking records, because preread operations can be avoided when  writing
; records.

b_write:	ld	a,c					; Set write operation given by register C.
		ld	(b_wrop),a				; Save it for BIOS use.
		ld	hl,(b_dmaadd)				; BIOS sector buffer.
		ld	de,b_dbuff				; MMC sector buffer.
		ld	bc,0080h				; Move a total of 128 bytes.
		ldir						; Move data.
		call	b_rwsetup
		call	b_MMCWriteSec
		jr	nc,b_w2
b_w1:		ld	a,0FFh					; Return error condition.
		ret
b_w2:		xor	a					; Return no error condition.
		ret

; LISTST Routine.

; You return the ready status of the list device used by the DESPOOL program to improve console response  during
; its operation. The value 0x00 is returned in A if the list device  is not ready to accept a character and 0xff
; if a character can be sent to the printer. A 0x00 value should be returned if LIST status is not implemented.

b_listst:	xor	a					; Return list status of 0x00 (not ready).
		ret						; Done.

; SECTRN Routine.

; Logical-to-physical sector translation  is performed to  improve the overall  response of CP/M.  Standard CP/M
; systems are shipped with a skew factor of 6, where six physical sectors are skipped between each logical  read
; operation. This skew factor allows enough time between sectors for most programs to load their buffers without
; missing the next sector. In particular computer systems that use fast processors, memory, and disk subsystems,
; the skew  factor might  be changed  to improve  overall response.  However, the  user should maintain a single
; -density IBM-compatible version of CP/M for information transfer into and out of the computer system, using  a
; skew factor of 6.

; In general, SECTRAN receives a logical sector number relative  to zero in BC and a translate table address  in
; DE. The sector number is used as an index into the translate table, with the resulting physical sector  number
; in HL. For standard systems, the table and indexing code is provided in the CBIOS and need not be changed.

b_sectrn:	dec	bc					; Make sector number in BC zero based.
		ld	hl,bc					; Pass BC back in HL.
		ret						; Done.

; End of required BIOS routines.

; Start of general BIOS support and debug routines.

; --------------
;
; FUNCTION: Transmit a NULL terminated string of characters in HL.
;    ENTRY: HL
;     EXIT: N/A
; DESTROYS: NONE

b_TXstringHL:	ld	a,(hl)					; Grab the character.
		or	a					; Check for null terminator.
		ret	z					; Yes, we've finished so return to sender. ;o)
		ld	c,a					; Load C with the character to display.
		call	b_conout				; And display it.
		inc	hl					; Go to next character in string.
		jr	b_TXstringHL				; And process it.

		INCLUDE	"eZ80-CPM-MMC.inc"			; CP/M MMC routines.

; BIOS data and variable storage.

b_maxdrvs:	EQU	02h					; Number of drives in system.

; XLT    Address of the logical-to-physical translation vector, if used for this particular drive, or the  value
;        0x0000 if no sector translation takes place (that is, the  physical and logical sector numbers  are the
;        same). Disk drives with identical sector skew factors (skf) share the same translate tables.
; 0000   Scratch pad values for use within the BDOS, initial value is unimportant.
; 0000   Scratch pad values for use within the BDOS, initial value is unimportant.
; 0000   Scratch pad values for use within the BDOS, initial value is unimportant.
; DIRBUF Address of a 128 byte scratch pad area for directory operations within BDOS. All DPHs address the  same
;        scratch pad area.
; DPB    Address of a disk  parameter block for this drive.  Drives with identical disk characteristics  address
;        the same disk parameter block.
; CSV    Address of a scratch pad area used for software check for changed disks. This address is different  for
;        each DPH.
; ALV    Address  of a  scratch pad  area used  by the  BDOS to  keep disk  storage allocation information. This
;        address is different for each DPH.

; Disk Parameter Header for disk 0 and 1.

b_dpbase:	EQU	$
b_dpe0:		DW	b_xlt0,0000h,0000h,0000h,b_dirbuf,b_dpb0,b_csv0,b_alv0
b_dpe1:		DW	b_xlt1,0000h,0000h,0000h,b_dirbuf,b_dpb0,b_csv1,b_alv1

; Disk Parameter Block for drive 0. Calculated using DISKDEF values 0,1,32,0,4096,2048,64,0,0

b_dpb0:		DW	32					; SPT total number of sectors per track.
		DB	5					; BSH data allocation block shift factor, determined by the data block allocation size.
		DB	31					; BLM data allocation block mask (2[BSH - 1]).
		DB	1					; EXM extent mask determined by the data block allocation size and the number of disk blocks.
		DW	2047					; DSM determines total storage capacity of this drive. (disk size - 1).
		DW	63					; DRM total number of directory entries that can be stored on this drive.
		DB	128					; AL0 determine reserved directory blocks.
		DB	0					; AL1 determine reserved directory blocks.
		DW	0					; CKS size of the directory check vector.
		DW	0					; OFF number of reserved tracks at the beginning of the disk.
b_xlt0:		EQU	0
b_csv0:		DS	1					; Check vector for drive 0.
b_alv0:		DS	256					; Allocation vector for drive 0.

; Disk Parameter Block for drive 1. Calculated using DISKDEF values 1,1,32,0,8192,1024,64,0,0

b_dpb1:		DW	32					; SPT total number of sectors per track.
		DB	6					; BSH data allocation block shift factor, determined by the data block allocation size.
		DB	63					; BLM data allocation block mask (2[BSH - 1]).
		DB	3					; EXM extent mask determined by the data block allocation size and the number of disk blocks.
		DW	1023					; DSM determines total storage capacity of this drive. (disk size - 1).
		DW	63					; DRM total number of directory entries that can be stored on this drive.
		DB	128					; AL0 determine reserved directory blocks.
		DB	0					; AL1 determine reserved directory blocks.
		DW	0					; CKS size of the directory check vector.
		DW	0					; OFF number of reserved tracks at the beginning of the disk.
b_xlt1:		EQU	0
b_csv1:		DS	1					; Check vector for drive 1.
b_alv1:		DS	128					; Allocation vector for drive 1.

; BIOS data storage.

b_wrop:		DB	00h					; Current write operation as passed from BDOS.
b_diskno:	DB	00h					; Current disk number as passed from BDOS.
b_track:	DW	0000h					; Current track number as passed from BDOS.
b_sector:	DW	0000h					; Current sector number as passed from BDOS.
b_dmaadd:	DW	0000h					; Current DMA address as passed from BDOS.


b_bootmsg:	ASCII	"\r\neZ80F91 Acclaim! 512k CP/M vers 2.2 (c) (p) 1982 Digital Research Inc."
		ASCII	"\r\nBIOS vers 1.2 2004 by www.vegeneering.com"
		ASCIZ	"\r\n\r\nMMC FLASH drive: "	     
		
b_InitSucc:	ASCIZ	"ENABLED.\r\n"

b_InitFail:	ASCIZ	"FAILED ** SYSTEM HALTED **\r\n"

		DS	20h					; Start of BIOS stack area.
b_biosstack:	EQU	$

b_dirbuf:	DS	80h					; 128 byte BDOS scratch pad.

b_dbuff:	DS	200h					; 512 byte MMC sector buffer.
