
; BOOT - Boot CP/M from the eZ80 on-chip FLASH.

; Configure the eZ80F91 modules memory.

; SRAM refers to the 512kB of Static RAM used on the eZ80F91 module (IC U8).
; FLASH refers to the 256kB of FLASH memory inside the eZ80F91 chip (IC U5).

; The following code will:

; a) Configure and enable SRAM to address range 0xC00000-0xC7FFFF.
; b) Copy the CP/M image code from FLASH to SRAM in ADL mode (24bit addressing).
; c) Reconfigure SRAM to address range 0x000000-0x07FFFF.
; d) Disable the FLASH and the SRAM should then become active.

; 16kB SRAM not used so we disable it.SUITE SORROW

start:		ld	a,00h					; Disable on-chip SRAM, Bit 7 GPRAM, it 6 EMACRAM.
		out0	(RAM_CTL),a

; CS0 1MB FLASH not used so we disable it. 040000h-11FFFFh

		ld	a,0E0h					; Bit 3 controls chip select.
		out0	(CS0_CTL),a

; CS1 512kB SRAM.

		ld	a,CodeStartADL >> 16			; Start at 0xC00000.
		out0	(CS1_LBR),a
		ld	a,CodeStartADL >> 16 + 7		; Ends  at 0xC7FFFF.
		out0	(CS1_UBR),a
		ld	a,028h					; Bit 7:5 1 wait state, bit 3 controls chip select.
		out0	(CS1_CTL),a

; Copy our code from FLASH to SRAM in 24 bit mode by setting ADL to 1.

		.ASSUME	ADL = 1					; Needed?

		stmix						; Needed?

		ld.lil	hl,mycodeS
		ld.lil	de,mycodeD
		ld.lil	bc,mycodeL
		ldir.lil

		rsmix						; Needed?

		.ASSUME	ADL = 0					; Needed?

; CS1 512kB SRAM.

		ld	a,00h					; Start at 0x000000.
		out0	(CS1_LBR),a
		ld	a,07h					; Ends  at 0x07FFFF.
		out0	(CS1_UBR),a
		ld	a,028h					; Bit 7:5 1 wait state, bit 3 controls chip select.
		out0	(CS1_CTL),a

; CS2 CF Interface. Currently not used so we disable it.

		ld	a,0A1h					; Start at 0xA10000.
		out0	(CS2_LBR),a
		ld	a,0A2h					; Ends  at 0xA2FFFF.
		out0	(CS2_UBR),a
		ld	a,0E0h					; Bit 3 controls chip select.
		out0	(CS2_CTL),a				; Added major wait states here just incase.

; CS3 not used so we disable it.

		ld	a,0E0h					; Bit 3 controls chip select.
		out0	(CS3_CTL),a

; Disable FLASH.

		ld	a,80h					; Reset = 88h. Bit 3 enables/disables FLASH.
		out0	(FLASH_CTRL),a

; At this point here, the FLASH memory should be disabled and the SRAM should be active
; and the PC should be pointing to the next instruction which is the jump to the BIOS boot routine.

