
; MMC - MMC Interface routines.

; Based on C source code by
; Copyright (c) 
; Revely Microsystems, 2003.
; http://www.revely.com/
;
; Please read the Revely License Agreement at
; http://www.revely.com/Source%20Code/readme.txt

; Converted to Z80 assembler and modified to work on the 
; eZ80F91 platform by 
; Jean-Michel Howland
; vegeneering, 2004.
; http://www.vegeneering.com/

GO_IDLE_STATE:		EQU	0
SEND_OP_COND:		EQU	1
READ_SINGLE_BLOCK:	EQU	17
WRITE_BLOCK:		EQU	24

; --------------

b_MMCInit:	call	b_SPIInit
		ld	b,0Ah					; Setup to loop 10 times.
b_mi1:		call	b_SPICardDelay				; Send 80 dummy clocks (no CS).
		djnz	b_mi1					; Loop until B = 0.
		call	b_SPILowerCS				; Activate MMC.
		ld	a,GO_IDLE_STATE				; Set to SPI Mode.
		ld	hl,0000h
		ld	de,0000h
		call	b_MMCSendCMD
		call	b_SPICardDelay
		ld	b,14h					; Setup to loop 20 times.
b_mi2:		call	b_SPIGetByte
		cp	01h
		jr	z,b_mi3
		call	b_SPICardDelay
		djnz	b_mi2					; Loop until B = 0.
		scf						; Exit with an error.
		ret
b_mi3:		call	b_SPIRaiseCS
		call	b_SPICardDelay
b_mi4:		call	b_SPILowerCS
		ld	a,SEND_OP_COND
		ld	hl,0000h
		ld	de,0000h
		call	b_MMCSendCMD
		call	b_SPIGetR1Resp
		push	af
		call	b_SPIRaiseCS
		call	b_SPICardDelay
		pop	af
		cp	00h
		jr	nz,b_mi4								
		scf
		ccf
		ret

; --------------

b_MMCReadSec:	call	b_SPILowerCS
		ld	a,READ_SINGLE_BLOCK
		call	b_MMCSendCMD
		call	b_SPIGetByte
		call	b_SPIGetR1Resp
		cp	00h
		jr	nz,b_mrs4
b_mrs1:		call	b_SPICardDelay
		call	b_SPIGetByte
		cp	0FEh					; Check for ready to send data token.
		jr	nz,b_mrs1				; Loop until MMC is ready.
		ld	hl,b_dbuff
		ld	b,20h					; Setup to loop 32 times.
b_mrs2:		ld	d,b
		ld	b,10h					; Setup to loop 16 times.
b_mrs3:		call	b_SPICardDelay
		call	b_SPIGetByte				; Read byte from MMC.
		ld	(hl),a
		inc	hl
		djnz	b_mrs3					; Loop until B = 0.
		ld	b,d
		djnz	b_mrs2					; Loop until B = 0.
		call	b_SPICardDelay
		call	b_SPICardDelay
		call	b_SPIRaiseCS
		call	b_SPICardDelay
		scf
		ccf
		ret
b_mrs4:		call	b_SPIRaiseCS
		call	b_SPICardDelay
		scf
		ret

; --------------
 
b_MMCWriteSec:	call	b_SPILowerCS
		ld	a,WRITE_BLOCK
		call	b_MMCSendCMD
		call	b_SPIGetR1Resp
		cp	00h
		jr	nz,b_mws5
		call	b_SPICardDelay
		ld	a,0FEh
		call	b_SPISendByte				; Data token.
		ld	hl,b_dbuff
		ld	b,20h					; Setup to loop 32 times.
b_mws1:		ld	d,b
		ld	b,10h					; Setup to loop 16 times.
b_mws2:		ld	a,(hl)
		call	b_SPISendByte
		call	b_SPIGetByte
		inc	hl
		djnz	b_mws2					; Loop until B = 0.
		ld	b,d
		djnz	b_mws1					; Loop until B = 0.
		call	b_SPICardDelay				; Send CRC byte.
		call	b_SPICardDelay				; Send CRC byte.
		call	b_SPICardDelay
		call	b_SPIGetByte				; Data response token.
		push	af
b_mws3:		call	b_SPICardDelay
		call	b_SPIGetByte
		cp	00h
		jr	z,b_mws3				; While busy, loop.
		call	b_SPIRaiseCS
		call	b_SPICardDelay
		call	b_SPILowerCS
		call	b_SPICardDelay
		call	b_SPIGetByte
b_mws4:		call	b_SPICardDelay
		call	b_SPIGetByte
		cp	00h
		jr	z,b_mws4				; While busy, loop.
		call	b_SPIRaiseCS
		pop	af
		cp	0E5h					; Data token OK.
		jr	nz,b_mws6				; Error exit.
		scf
		ccf
		ret
b_mws5:		call	b_SPIRaiseCS
		call	b_SPICardDelay
b_mws6:		scf
		ret

; --------------

b_SPIRaiseCS:	in0	a,(PB_DR)
		set	0,a					; xxxx xxx1 Set bit 0 high.
		jr	b_srlcs1
b_SPILowerCS:	in0	a,(PB_DR)
		res	0,a					; xxxx xxx0 Set bit 0 low.
b_srlcs1:	out0	(PB_DR),a
		ret

; --------------

b_MMCSendCMD:	or	40h					; Set command bit.
		call	b_SPISendByte
		call	b_SL32					; Address must be aligned to 512 byte boundary.
		ld	a,h
		call	b_SPISendByte
		ld	a,l
		call	b_SPISendByte
		ld	a,d
		call	b_SPISendByte
		ld	a,e
		call	b_SPISendByte
		ld	a,95h
		call	b_SPISendByte
		ret

; --------------

b_SPIGetR1Resp:	call	b_SPIGetByte
		push	af
		call	b_SPICardDelay
		pop	af
		bit	7,a					; bit 7 = 1, z = 0, loop.
		jr	nz,b_SPIGetR1Resp
		ret

; --------------

b_SPIStatus:	push	af
b_ss1:		in0	a,(SPI_SR)
		and	80h
		jr	nz,b_ss1				; Loop until bit 7 = 1.
		pop	af
		ret

; --------------

b_SPICardDelay:	ld	a,0FFh
b_SPISendByte:	call	b_SPIStatus
		out0	(SPI_TSR),a
		ret

; --------------

b_SPIGetByte:	call	b_SPIStatus
		in0	a,(SPI_RBR)
		ret

; --------------

b_SPIInit:	ld	a,10001001b				; 89h.
		out0	(PB_DR),a
		ld	a,11001100b				; 0CCh.
		out0	(PB_DDR),a
		ld	a,00000000b				; 00h.
		out0	(PB_ALT1),a
		ld	a,11001100b				; 0CCh   Configure PORT B for alternative use.
		out0	(PB_ALT2),a				; bit 7O=MOSI,6I=MISO,5x=x,4x=x,3O=SCK,2I=SS,1x=x,0O=/CS.
		ld	a,00H					; Set SPI baudrate. SYSTEMCLOCK(in MHz) / (2 * ClockRate).
		out0	(SPI_BRG_H),a
		ld	a,04h					; @ 50MHz clock, 04h = ~6.25Mbit/s.
		out0	(SPI_BRG_L),a
		ld	a,30h					; 0011 0100 Enable SPI interface. CLK idles low.
		out0	(SPI_CTL),a
		ret

; --------------
;
; 32 bit shift left. 
;    ENTRY: N/A
;     EXIT: N/A
; DESTROYS: AF, B, IX

b_SL32:		ld	b,9					; Shift 32 bit number in HLDE left 9 bits.
		ld	ix,b_HLDE				; 4 bytes of memory.
		ld	(ix+0),h
		ld	(ix+1),l
		ld	(ix+2),d
		ld	(ix+3),e
b_sl321:	sla	(ix+0)
		sla	(ix+1)
		jr	nc,b_sl322
		set	0,(ix+0)
b_sl322:	sla	(ix+2)
		jr	nc,b_sl323
		set	0,(ix+1)
b_sl323:	sla	(ix+3)
		jr	nc,b_sl324
		set	0,(ix+2)
b_sl324:	djnz	b_sl321
		ld	h,(ix+0)
		ld	l,(ix+1)
		ld	d,(ix+2)
		ld	e,(ix+3)
		ret

b_HLDE:		DS	4

; End of MMC routines.
